from exectools import *


if __name__ == '__main__':
    # ExecExport("~/Desktop/test_exec_tools_empty.xlsx").to_excel([dict(No=111, display_name="123")])
    for i in ExecImport("~/Desktop/test_exec_tools.xlsx").from_excel():
        print(i)



import pandas as pd


f_path = '~/Desktop/test_exec_tools.xlsx'

data1 = pd.ExcelFile(f_path).sheet_names
# data2 = pd.read_excel(f_path)
# all_data = [data1, data2]

# data3 = pd.concat(all_data)
print(data1)



from exectools import RunTime
import time


@RunTime.run_time(show_short_result=10)
def test_run_time_res(sleep=1):
    time.sleep(sleep)
    return [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]


@RunTime.loop_time(show_short_result=10)
def test_loop_while_res(sleep=1):
    time.sleep(sleep)
    return [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]


@RunTime.loop_time(loop_times=2, show_short_result=10)
def test_loop_range_res(sleep=1):
    time.sleep(sleep)
    return [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]


@RunTime.timeout(6)
def test_timeout(sleep=5):
    time.sleep(sleep)
    return 'test timeout.'


if __name__ == '__main__':
    pass
    # test_run_time_res()
    # test_loop_while_res()
    # test_loop_range_res()
    try:
        print(test_timeout())
    except Exception as e:
        print(e)



from exectools.handler import ExecImport, ExecExport
from exectools.tools_basic import Tools, TimeSection
from exectools.tools_excel import ExecExcel
from exectools.tools_decorator import Retry, RunTime
from exectools.tools_coroutine import Coroutine


__all__ = [
    'ExecImport',
    'ExecExport',
    'ExecExcel',
    'Tools',
    'TimeSection',
    'Retry',
    'RunTime',
    'Coroutine',
]



from typing import List


class Coroutine:
    def __init__(self, handler, batch_size=10, sleep=0):
        self.__handler = handler
        self.batch_size = batch_size
        self.sleep = sleep

    def run_all(self, func, func_args: List[dict], batch_size=None, sleep=None, raise_error=True, **kwargs):
        for i in self.item_range(func_args, batch_size):
            workers = [self.__handler.spawn_later(sleep or self.sleep, func, **j) for j in func_args[i[0]:i[1]]]
            self.__handler.joinall(workers, raise_error=raise_error, **kwargs)
            yield [res.value for res in workers]

    def item_range(self, items: list, batch_size=None):
        batch_siez = batch_size or self.batch_size
        for i in range(0, len(items), batch_siez):
            cursor = i
            yield cursor, cursor + batch_siez



from exectools.tools_excel import ExecExcel
from exectools.tools_basic import Tools
from json import load, dump


class ExecImport:
    def __init__(self, path):
        self.path = Tools.extend_path(path)

    def from_json(self):
        with open(self.path, 'r') as f:
            return load(f)

    def from_excel(self, sheet='Sheet1', axis=0, index_col=None,
                   orient='records', empty=False, nana=None, **kwargs) -> iter:
        return ExecExcel(self.path).read(sheet, axis, index_col, **kwargs).extract_data(orient, empty, nana)


class ExecExport:
    def __init__(self, path):
        self.path = Tools.extend_path(path)

    def to_json(self, data):
        with open(self.path, 'w') as f:
            dump(data, f)

    def to_excel(self, data: list, sheet='Sheet1', file_path=None, index=False, axis=0, index_col=None, **kwargs):
        ExecExcel(self.path).append_row(data, sheet, axis, index_col).to_excel(sheet, file_path, index, **kwargs)



from gevent import monkey; monkey.patch_all()
from exectools import Coroutine
import gevent


def count_num(a, b):
    print(f"Start: {a}")
    gevent.sleep(5)
    return a/b


if __name__ == '__main__':
    pass
    items_ = [dict(a=2, b=2), dict(a=4, b=2), dict(a=1, b=0), dict(a=6, b=2)]
    items1 = [1, 2, 3]
    try:
        for i in Coroutine(gevent).run_all(count_num, items_, batch_size=2):
            print(i)
    except Exception as e:
        print('error')
        print(e)



from os.path import exists, expanduser
import pandas as pd
import time


class ExecExcel:
    """
    read xlsx and csv
    """
    def __init__(self, file_path):
        self.file_path = expanduser(file_path)

    def read(self, sheet='Sheet1', axis=0, index_col=None, **kwargs):
        df = pd.ExcelFile(self.file_path)
        sheets = [sheet] if sheet else df.sheet_names
        df_parse = df.parse(sheets, index_col=index_col, **kwargs)
        frame_data = pd.concat(df_parse, axis=axis)
        return ExcelResponse(frame_data, self.file_path)

    def read_csv(self, **kwargs):
        frame_data = pd.read_csv(self.file_path, **kwargs)
        return ExcelResponse(frame_data, self.file_path)

    def data_format(self, data: list, axis=0):
        """
        Write data to excel.
        :param axis:
        :param data: dict in list
        """
        fd = [pd.DataFrame(item, index=[0]) for item in data]
        frame_data = pd.concat(fd, axis=axis)
        return ExcelResponse(frame_data, self.file_path)

    def append_row(self, data: list, sheet='Sheet1', axis=0, index_col=None, **kwargs):
        if exists(self.file_path):
            df = pd.ExcelFile(self.file_path)
            sheets = [sheet] if sheet else df.sheet_names
            df_parse = df.parse(sheets, index_col=index_col, **kwargs)
            frame_data = pd.concat(df_parse, axis=axis)
        else:
            frame_data = pd.DataFrame()
        new_data = pd.concat([pd.DataFrame(item, index=[0]) for item in data], axis=axis)
        frame_data = pd.concat([frame_data, new_data], axis=axis)
        return ExcelResponse(frame_data, self.file_path)


class ExcelResponse:
    def __init__(self, frame_data, file_path: str):
        self.frame_data = frame_data
        self.file_path = file_path

    def extract_data(self, orient='records', empty=False, nana=None):
        """
        :param empty: if True show empty value
        :param orient:
        {
        orient=’records’: [{column -> value}, … , {column -> value}],
        orient=’index’  : {index -> {column -> value}},
        orient=’dict’   : {column -> {index -> value}},
        orient=’list’   : {column -> [values]} ,
        orient=’series’ : {column -> Series(values)},
        orient=’split’  : {index -> [index], columns -> [columns], data -> [values]},
        }
        :param nana: null display NaNa/None/''
        :return: dict : generator object
        """
        frame_data = self.frame_data.where(self.frame_data.notnull(), nana)
        data_load = frame_data.to_dict(orient=orient)
        if orient == 'records':
            for index, value in enumerate(data_load):
                yield {k: v for k, v in value.items()} if empty else {k: v for k, v in value.items() if v}
        else:
            for index, value in data_load.items():
                if value is dict:
                    yield index, {k: v for k, v in value.items()} if empty else {k: v for k, v in value.items() if v}
                else:
                    yield {index: value}

    def to_excel(self, sheet="Sheet1", file_path=None, index=False, **kwargs):
        """
        Write data to excel.
        :param index:
        :param file_path: save excel file name
        :param sheet: excel sheet name
        """
        file_path = file_path or self.file_path.rsplit('.')[0] + f'_{int(time.time())}.xlsx'
        self.frame_data.to_excel(file_path, index=index, sheet_name=sheet, **kwargs)



import functools
import threading
import ctypes
import time


class TerminableThread(threading.Thread):
    """a thread that can be stopped by forcing an exception in the execution context"""

    def terminate(self, exception_cls, repeat_sec=2.0):
        if self.is_alive() is False:
            return True
        killer = ThreadKiller(self, exception_cls, repeat_sec=repeat_sec)
        killer.start()


class ThreadKiller(threading.Thread):
    """separate thread to kill TerminableThread"""

    def __init__(self, target_thread, exception_cls, repeat_sec=2.0):
        threading.Thread.__init__(self)
        self.target_thread = target_thread
        self.exception_cls = exception_cls
        self.repeat_sec = repeat_sec
        self.daemon = True

    def run(self):
        """loop raising exception incase it's caught hopefully this breaks us far out"""
        while self.target_thread.is_alive():
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self.target_thread.ident),
                                                       ctypes.py_object(self.exception_cls))
            self.target_thread.join(self.repeat_sec)


class Retry:
    @classmethod
    def trying(cls, retry_times: int = 3, retry_sleep: int = 0, retry_exceptions: tuple = None, logger=print,
               sleep_func=time.sleep):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except (retry_exceptions or Exception) as e:
                    return cls.__retry(func, retry_times, e, retry_sleep, logger, sleep_func, *args, **kwargs)
            return wrapper
        return decorator

    @classmethod
    def __retry(cls, func, retry_times, error, retry_sleep, logger, sleep_func=time.sleep, *args, **kwargs):
        if retry_times >= 1:
            try:
                retry_times -= 1
                logger(f'Retry remaining times: {retry_times}; Will sleep {retry_sleep}s; Error: {error};')
                sleep_func(retry_sleep)
                return func(*args, **kwargs)
            except Exception as e:
                return cls.__retry(func, retry_times, e, retry_sleep, logger, sleep_func, *args, **kwargs)
        else:
            raise


class RunTime:
    @classmethod
    def run_time(cls, logger=print, show_short_result: int = None):
        """
        :param logger: logger func, default is python print
        :param show_short_result: displays the result of the number of characters.
        """
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                start = time.time()
                res = func(*args, **kwargs)
                end = time.time()
                str_res = f'{res}'
                logger(
                    f'\n##########################################'
                    f'\nFUNC: {func.__name__}'
                    f'\nSTART: {round(start, 2)} END: {round(end, 2)}'
                    f'\nTOTAL TIME: {round(end - start, 2)}'
                    f'\n##########################################'
                    f'\nRESULT: '+(str_res[:show_short_result]+'......'
                                   if show_short_result and len(str_res) > show_short_result else str_res))
                return res
            return wrapper
        return decorator

    @classmethod
    def loop_time(cls, loop_times: int = None, loop_sleep: int = 0, logger=print, show_short_result: int = None,
                  sleep_func=time.sleep):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                n, flag = {'time': 1}, True
                start = time.time()
                try:
                    (cls.__rangetime(loop_times, n, func, loop_sleep, logger, show_short_result, sleep_func,
                                     *args, **kwargs)
                     if loop_times else
                     cls.__whiletime(flag, n, func, loop_sleep, logger, show_short_result, sleep_func, *args, **kwargs))

                except KeyboardInterrupt:
                    pass
                finally:
                    end = time.time()
                    n['time'] -= 1
                    logger(
                        f'\n##########################################'
                        f'\nFUNC: {func.__name__}'
                        f'\nSTART: {round(start, 2)} END: {round(end, 2)}'
                        f'\nTOTAL_TIME: {round(end - start, 2)} RUN_TIMES: {n["time"]}'
                        f'\n##########################################')
            return wrapper
        return decorator

    @classmethod
    def __rangetime(cls, times, n, func, sleep, logger=print, show_short_result: int = None, sleep_func=time.sleep,
                    *args, **kwargs):
        res = True
        for _ in range(times):
            if res:
                start = time.time()
                res = func(*args, **kwargs)
                end = time.time()
                str_res = f'{res}'
                show_res = (f'{res}'[:show_short_result]+'......'
                            if show_short_result and len(str_res) > show_short_result else f'{res}')
                logger(f'Times: {n["time"]}, Time_cost: {end - start}\n'
                       f'RESULT: {show_res or "Range loop: Nothing returned."}\n')
                sleep_func(sleep)
                n["time"] += 1
            else:
                break

    @classmethod
    def __whiletime(cls, flag, n, func, sleep, logger=print, show_short_result: int = None, sleep_func=time.sleep,
                    *args, **kwargs):
        while flag:
            start = time.time()
            res = func(*args, **kwargs)
            end = time.time()
            str_res = f'{res}'
            show_res = (f'{res}'[:show_short_result] + '......'
                        if show_short_result and len(str_res) > show_short_result else f'{res}')
            logger(f'Times: {n["time"]}, Time_cost: {end - start}\n'
                   f'RESULT: {show_res or "While loop: Nothing returned."}\n')
            sleep_func(sleep)
            n["time"] += 1

    @classmethod
    def timeout(cls, sec, raise_sec=1):
        """
        timeout decorator
        :param sec: function raise TimeoutError after ? seconds
        :param raise_sec: retry kill thread per ? seconds. default: 1 second
        """
        def decorator(func):
            @functools.wraps(func)
            def wrapped_func(*args, **kwargs):
                err_msg = f'Function {func.__name__} timed out after {sec} seconds'

                class FuncTimeoutError(TimeoutError):
                    def __init__(self):
                        TimeoutError.__init__(self, err_msg)

                result, exception = [], []

                def run_func():
                    try:
                        res = func(*args, **kwargs)
                    except FuncTimeoutError:
                        pass
                    except Exception as e:
                        exception.append(e)
                    else:
                        result.append(res)

                # typically, a python thread cannot be terminated, use TerminableThread instead
                thread = TerminableThread(target=run_func, daemon=True)
                thread.start()
                thread.join(timeout=sec)

                if thread.is_alive():
                    # a timeout thread keeps alive after join method, terminate and raise TimeoutError
                    exc = type('TimeoutError', FuncTimeoutError.__bases__, dict(FuncTimeoutError.__dict__))
                    thread.terminate(exception_cls=exc, repeat_sec=raise_sec)
                    raise TimeoutError(err_msg)
                elif exception:
                    # if exception occurs during the thread running, raise it
                    raise exception[0]
                else:
                    # if the thread successfully finished, return its results
                    return result[0]
            return wrapped_func
        return decorator



from exectools.tools_basic import *
from time import time


if __name__ == '__main__':
    print(datetime.now())
    print(TimeSection.nowstamp_point(time_type='hours', time_type_offset=1))
    print(TimeSection.timestamp_point(time(), "hours"))



from exectools.handler import *


class TestExecExcel:
    __excel_path = '~/Desktop/test_exec_tools.xlsx'
    __csv_path = '~/Desktop/test_exec_tools.csv'
    __excel_data = ExecExcel(__excel_path)
    __csv_data = ExecExcel(__csv_path)

    @classmethod
    def test_read(cls, sheet='Sheet1'):
        for i in cls.__excel_data.read(sheet=sheet).extract_data():
            print(i)

    @classmethod
    def test_read_csv(cls):
        for i in cls.__csv_data.read_csv().extract_data():
            print(i)

    @classmethod
    def test_read_to_excel(cls, sheet='Sheet1'):
        cls.__excel_data.read(sheet=sheet).to_excel()

    @classmethod
    def test_read_csv_to_excel(cls):
        cls.__csv_data.read_csv().to_excel()

    @classmethod
    def test_data_format(cls, data):
        for i in cls.__excel_data.data_format(data=data).extract_data():
            print(i)

    @classmethod
    def test_append_row(cls, data):
        for i in cls.__excel_data.append_row(data=data).extract_data():
            print(i)

    @classmethod
    def test_append_row_to_excel(cls, data):
        cls.__excel_data.append_row(data=data).to_excel()


if __name__ == '__main__':
    # TestExecExcel.test_append_row([dict(No=111, display_name="123")])
    TestExecExcel.test_append_row_to_excel([dict(No=111, display_name="123")])
    # TestExecExcel.test_read_to_excel(sheet="")



import setuptools
import re

pattern = r'^\#.*'
with open("README.md", "r") as fh:
    long_description = fh.read()

proj_name = re.search(pattern, long_description).group()[2:]
description = "Some Tools."
install_requires = ['pandas', 'openpyxl']
packages_include = ('exectools',)
author = "Supplayer"
author_email = "x254724521@hotmail.com"

setuptools.setup(
    name='-'.join(proj_name.split('_')) if '_' in proj_name else proj_name,
    author=author,
    author_email=author_email,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=f"https://github.com/{author}/{proj_name}.git",
    packages=setuptools.find_packages(include=packages_include),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=install_requires,
    setup_requires=['setuptools_scm'],
    use_scm_version=True
)



from os.path import expanduser
from datetime import datetime, timedelta


class Tools:
    @classmethod
    def extend_path(cls, path: str):
        return expanduser(path)


class TimeSection:
    __time_type = dict(year=1, month=2, days=3, hours=4, minutes=5, seconds=6)
    __datetime_format = "%Y:%m:%d:%H:%M:%S"

    @classmethod
    def nowstamp_point(cls, time_type_offset: int = 0, time_type: str = "hours", res_func=int):
        """
        :param time_type: days/hours/minutes/seconds
        :param time_type_offset: timedelta offset
        :param res_func: format timestamp
        :return: timestamp
        """
        time_type_offset = timedelta(**{time_type: time_type_offset})
        timestamp = (datetime.now() + time_type_offset).timestamp()
        return cls.timestamp_point(timestamp, time_type, res_func)

    @classmethod
    def timestamp_point(cls, time_stamp: float, time_type: str = "hours", res_func=int):
        point = cls.__time_type[time_type]
        now = datetime.fromtimestamp(time_stamp)
        date_string = ":".join(now.strftime(cls.__datetime_format).split(':')[:point]) + cls.__fill_format(point)
        return res_func(datetime.strptime(date_string, cls.__datetime_format).timestamp())

    @classmethod
    def __fill_format(cls, point: int):
        return ":00"*(len(cls.__time_type) - point)
